@@ -4,6 +4,7 @@
 #![allow(unused_braces)]
 #![allow(clippy::type_complexity)]
 #![allow(clippy::upper_case_acronyms)]
+#![allow(clippy::needless_lifetimes)]
 
 pub use internal::{Parser, Token};
 
@@ -14,8 +15,13 @@ pomelo! {
     %stack_size 0;
 
     %parser pub struct Parser<'a, T: visitors::Smt2Visitor> {};
-    %extra_argument &'a mut T;
-
+    %extra_argument (&'a mut T, &'a mut crate::lexer::Position);
+    %error T::Error;
+    %syntax_error { match token {
+        Some(token) => Err(extra.0.syntax_error(extra.1.clone(), format!("unexpected token: {:?}", token))),
+        None => Err(extra.0.syntax_error(extra.1.clone(), "unexpected end of input".into())),
+    }};
+    %parse_fail { extra.0.parsing_error(extra.1.clone(), "unrecoverable parsing error".into()) };
     %token #[derive(Clone, Debug, PartialEq, Eq)] pub enum Token {};
 
     %type Keyword String;
@@ -86,10 +92,10 @@ pomelo! {
 
     %start_symbol command;
 
-    bound_symbol ::= Symbol(s) { extra.visit_bound_symbol(s).expect("failed to resolve symbol") }
-    fresh_symbol ::= Symbol(s) { extra.visit_fresh_symbol(s) }
-    any_symbol ::= Symbol(s) { extra.visit_any_symbol(s) }
-    keyword ::= Keyword(s) { extra.visit_keyword(s) }
+    bound_symbol ::= Symbol(s) { extra.0.visit_bound_symbol(s)? }
+    fresh_symbol ::= Symbol(s) { extra.0.visit_fresh_symbol(s)? }
+    any_symbol ::= Symbol(s) { extra.0.visit_any_symbol(s)? }
+    keyword ::= Keyword(s) { extra.0.visit_keyword(s)? }
 
     bound_symbols ::= bound_symbol(x) { vec![x] }
     bound_symbols ::= bound_symbols(mut xs) bound_symbol(x) { xs.push(x); xs }
@@ -108,10 +114,10 @@ pomelo! {
     attributes ::= attributes(mut xs) keyword(k) attribute_value(v) { xs.push((k, v)); xs }
 
     // s_expr ::= ⟨spec_constant⟩ | ⟨symbol⟩ | ⟨keyword⟩ | ( ⟨s_expr⟩∗ )
-    s_expr ::= constant(x) { extra.visit_constant_s_expr(x) }
-    s_expr ::= bound_symbol(x) { extra.visit_symbol_s_expr(x) }
-    s_expr ::= keyword(x) { extra.visit_keyword_s_expr(x) }
-    s_expr ::= LeftParen s_exprs?(xs) RightParen { extra.visit_application_s_expr(xs.unwrap_or_else(Vec::new)) }
+    s_expr ::= constant(x) { extra.0.visit_constant_s_expr(x)? }
+    s_expr ::= bound_symbol(x) { extra.0.visit_symbol_s_expr(x)? }
+    s_expr ::= keyword(x) { extra.0.visit_keyword_s_expr(x)? }
+    s_expr ::= LeftParen s_exprs?(xs) RightParen { extra.0.visit_application_s_expr(xs.unwrap_or_else(Vec::new))? }
 
     s_exprs ::= s_expr(x) { vec![x] }
     s_exprs ::= s_exprs(mut xs) s_expr(x) { xs.push(x); xs }
@@ -128,22 +134,22 @@ pomelo! {
     identifier ::= LeftParen Underscore bound_symbol(symbol) indices(indices) RightParen { visitors::Identifier::Indexed { symbol, indices } }
 
     // sort ::= ⟨identifier⟩ | ( ⟨identifier⟩ ⟨sort⟩+ )
-    sort ::= identifier(id) { extra.visit_simple_sort(id) }
-    sort ::= LeftParen identifier(id) sorts(sorts) RightParen { extra.visit_parameterized_sort(id, sorts) }
+    sort ::= identifier(id) { extra.0.visit_simple_sort(id)? }
+    sort ::= LeftParen identifier(id) sorts(sorts) RightParen { extra.0.visit_parameterized_sort(id, sorts)? }
 
     sorts ::= sort(x) { vec![x] }
     sorts ::= sorts(mut xs) sort(x) { xs.push(x); xs }
 
     // qual_identifier ::= ⟨identifier⟩ | ( as ⟨identifier⟩ ⟨sort⟩ )
-    qual_identifier ::= identifier(x) { extra.visit_simple_identifier(x) }
-    qual_identifier ::= LeftParen As identifier(id) sort(s) RightParen { extra.visit_sorted_identifier(id, s) }
+    qual_identifier ::= identifier(x) { extra.0.visit_simple_identifier(x)? }
+    qual_identifier ::= LeftParen As identifier(id) sort(s) RightParen { extra.0.visit_sorted_identifier(id, s)? }
 
     // constant ::= ⟨numeral⟩ | ⟨decimal⟩ | ⟨hexadecimal⟩ | ⟨binary⟩ | ⟨string⟩
-    constant ::= Numeral(x) { extra.visit_numeral_constant(x) }
-    constant ::= Decimal(x) { extra.visit_decimal_constant(x) }
-    constant ::= Hexadecimal(x) { extra.visit_hexadecimal_constant(x) }
-    constant ::= Binary(x) { extra.visit_binary_constant(x) }
-    constant ::= String(x) { extra.visit_string_constant(x) }
+    constant ::= Numeral(x) { extra.0.visit_numeral_constant(x)? }
+    constant ::= Decimal(x) { extra.0.visit_decimal_constant(x)? }
+    constant ::= Hexadecimal(x) { extra.0.visit_hexadecimal_constant(x)? }
+    constant ::= Binary(x) { extra.0.visit_binary_constant(x)? }
+    constant ::= String(x) { extra.0.visit_string_constant(x)? }
 
     // ⟨var_binding⟩ ::= ( ⟨symbol⟩ ⟨term⟩ )
     var_binding ::= LeftParen fresh_symbol(s) term(t) RightParen { (s, t) }
@@ -169,21 +175,21 @@ pomelo! {
 
     // terms ::= ...
     //   ⟨spec_constant⟩
-    term ::= constant(x) { extra.visit_constant(x) }
+    term ::= constant(x) { extra.0.visit_constant(x)? }
     //   ⟨qual_identifier⟩
-    term ::= qual_identifier(x) { extra.visit_qual_identifier(x) }
+    term ::= qual_identifier(x) { extra.0.visit_qual_identifier(x)? }
     //   ( let ( ⟨var_binding⟩+ ) ⟨term⟩ )
-    term ::= LeftParen Let LeftParen var_bindings(xs) RightParen term(t) RightParen { extra.visit_let(xs, t) }
+    term ::= LeftParen Let LeftParen var_bindings(xs) RightParen term(t) RightParen { extra.0.visit_let(xs, t)? }
     //   ( forall ( ⟨sorted_var⟩+ ) ⟨term⟩ )
-    term ::= LeftParen Forall LeftParen sorted_vars(xs) RightParen term(t) RightParen { extra.visit_forall(xs, t) }
+    term ::= LeftParen Forall LeftParen sorted_vars(xs) RightParen term(t) RightParen { extra.0.visit_forall(xs, t)? }
     //   ( exists ( ⟨sorted_var⟩+ ) ⟨term⟩ )
-    term ::= LeftParen Exists LeftParen sorted_vars(xs) RightParen term(t) RightParen { extra.visit_exists(xs, t) }
+    term ::= LeftParen Exists LeftParen sorted_vars(xs) RightParen term(t) RightParen { extra.0.visit_exists(xs, t)? }
     //   ( match ⟨term⟩ ( ⟨match_case⟩+ ) )
-    term ::= LeftParen Match term(t) LeftParen match_cases(xs) RightParen RightParen { extra.visit_match(t, xs) }
+    term ::= LeftParen Match term(t) LeftParen match_cases(xs) RightParen RightParen { extra.0.visit_match(t, xs)? }
     //   ( ! ⟨term⟩ ⟨attribute⟩+ )
-    term ::= LeftParen Exclam term(t) attributes(xs) RightParen { extra.visit_attributes(t, xs) }
+    term ::= LeftParen Exclam term(t) attributes(xs) RightParen { extra.0.visit_attributes(t, xs)? }
     //   ( ⟨qual_identifier ⟩ ⟨term⟩+ )
-    term ::= LeftParen qual_identifier(id) terms(xs) RightParen { extra.visit_application(id, xs) }
+    term ::= LeftParen qual_identifier(id) terms(xs) RightParen { extra.0.visit_application(id, xs)? }
 
     terms ::= term(x) { vec![x] }
     terms ::= terms(mut xs) term(x) { xs.push(x); xs }
@@ -192,7 +198,7 @@ pomelo! {
     prop_literal ::= bound_symbol(x) { (x, true) }
     prop_literal ::= LeftParen Symbol(s) bound_symbol(x) RightParen {
         if s != "not" {
-            return Err(());
+            return Err(extra.0.parsing_error(extra.1.clone(), format!("invalid prop_literal: found {} instead of `not`", s)));
         }
         (x, false)
     }
@@ -249,23 +255,23 @@ pomelo! {
 
     // command ::= ...
     //   ( assert ⟨term⟩ )
-    command ::= LeftParen Assert term(t) RightParen { extra.visit_assert(t) }
+    command ::= LeftParen Assert term(t) RightParen { extra.0.visit_assert(t)? }
     //   ( check-sat )
-    command ::= LeftParen CheckSat RightParen { extra.visit_check_sat() }
+    command ::= LeftParen CheckSat RightParen { extra.0.visit_check_sat()? }
     //   ( check-sat-assuming ( ⟨prop_literal⟩∗ ) )
     command ::= LeftParen CheckSatAssuming LeftParen prop_literals?(xs) RightParen RightParen
     {
-        extra.visit_check_sat_assuming(xs.unwrap_or_else(Vec::new))
+        extra.0.visit_check_sat_assuming(xs.unwrap_or_else(Vec::new))?
     }
     //   ( declare-const ⟨symbol⟩ ⟨sort⟩ )
     command ::= LeftParen DeclareConst fresh_symbol(x) sort(s) RightParen
     {
-        extra.visit_declare_const(x, s)
+        extra.0.visit_declare_const(x, s)?
     }
     //   ( declare-datatype ⟨symbol⟩ ⟨datatype_dec⟩)
     command ::= LeftParen DeclareDatatype fresh_symbol(s) datatype_dec(d) RightParen
     {
-        extra.visit_declare_datatype(s, d)
+        extra.0.visit_declare_datatype(s, d)?
     }
     //   ( declare-datatypes ( ⟨sort_dec⟩n+1 ) ( ⟨datatype_dec⟩n+1 ) )
     command ::= LeftParen DeclareDatatypes LeftParen sort_decs(sorts) RightParen LeftParen datatype_decs(datatypes) RightParen RightParen
@@ -276,82 +282,82 @@ pomelo! {
                 .zip(datatypes.into_iter())
                 .map(|((sort, arity), datatype)| (sort, arity, datatype))
                 .collect::<Vec<_>>();
-            extra.visit_declare_datatypes(results)
+            extra.0.visit_declare_datatypes(results)?
         } else {
-            return Err(());
+            return Err(extra.0.parsing_error(extra.1.clone(), format!("wrong number of types in `declare-datatypes`: {} instead of {}", datatypes.len(), sorts.len())));
         }
     }
     //   ( declare-fun ⟨symbol⟩ ( ⟨sort⟩∗ ) ⟨sort⟩ )
     command ::= LeftParen DeclareFun fresh_symbol(x) LeftParen sorts?(xs) RightParen sort(r) RightParen
     {
-        extra.visit_declare_fun(x, xs.unwrap_or_else(Vec::new), r)
+        extra.0.visit_declare_fun(x, xs.unwrap_or_else(Vec::new), r)?
     }
     //   ( declare-sort ⟨symbol⟩ ⟨numeral⟩ )
     command ::= LeftParen DeclareSort fresh_symbol(x) Numeral(num) RightParen
     {
-        extra.visit_declare_sort(x, num)
+        extra.0.visit_declare_sort(x, num)?
     }
     //   ( define-fun ⟨function_dec⟩ ⟨term⟩ )
     command ::= LeftParen DefineFun function_dec(d) term(x) RightParen
     {
-        extra.visit_define_fun(d, x)
+        extra.0.visit_define_fun(d, x)?
     }
     //   ( define-fun-rec ⟨function_dec⟩ ⟨term⟩ )
     command ::= LeftParen DefineFunRec function_dec(d) term(x) RightParen
     {
-        extra.visit_define_fun_rec(d, x)
+        extra.0.visit_define_fun_rec(d, x)?
     }
     //   ( define-funs-rec ( ( ⟨function_dec⟩ )n+1 ) ( ⟨term⟩n+1 ) )
     command ::= LeftParen DefineFunsRec LeftParen function_decs(sigs) RightParen LeftParen terms(xs) RightParen RightParen
     {
         if sigs.len() == xs.len() {
             let defs = sigs.into_iter().zip(xs.into_iter()).collect::<Vec<_>>();
-            extra.visit_define_funs_rec(defs)
+            extra.0.visit_define_funs_rec(defs)?
         } else {
-            return Err(());
+            return Err(extra.0.parsing_error(extra.1.clone(), format!("wrong number of function bodies in `define-funs-rec`: {} instead of {}", xs.len(), sigs.len())));
         }
     }
     //   ( define-sort ⟨symbol⟩ ( ⟨symbol⟩∗ ) ⟨sort⟩ )
     command ::= LeftParen DefineSort fresh_symbol(x) LeftParen bound_symbols?(xs) RightParen sort(r) RightParen
     {
-        extra.visit_define_sort(x, xs.unwrap_or_else(Vec::new), r)
+        extra.0.visit_define_sort(x, xs.unwrap_or_else(Vec::new), r)?
     }
     //   ( echo ⟨string⟩ )
-    command ::= LeftParen Echo String(x) RightParen { extra.visit_echo(x) }
+    command ::= LeftParen Echo String(x) RightParen { extra.0.visit_echo(x)? }
     //   ( exit )
-    command ::= LeftParen Exit RightParen { extra.visit_exit() }
+    command ::= LeftParen Exit RightParen { extra.0.visit_exit()? }
     //   ( get-assertions )
-    command ::= LeftParen GetAssertions RightParen { extra.visit_get_assertions() }
+    command ::= LeftParen GetAssertions RightParen { extra.0.visit_get_assertions()? }
     //   ( get-assignment )
-    command ::= LeftParen GetAssignments RightParen { extra.visit_get_assignment() }
+    command ::= LeftParen GetAssignment RightParen { extra.0.visit_get_assignment()? }
     //   ( get-info ⟨info_flag⟩ )
-    command ::= LeftParen GetInfo keyword(x) RightParen { extra.visit_get_info(x) }
+    command ::= LeftParen GetInfo keyword(x) RightParen { extra.0.visit_get_info(x)? }
     //   ( get-model )
-    command ::= LeftParen GetModel RightParen { extra.visit_get_model() }
+    command ::= LeftParen GetModel RightParen { extra.0.visit_get_model()? }
     //   ( get-option ⟨keyword⟩ )
-    command ::= LeftParen GetOption keyword(x) RightParen { extra.visit_get_option(x) }
+    command ::= LeftParen GetOption keyword(x) RightParen { extra.0.visit_get_option(x)? }
     //   ( get-proof )
-    command ::= LeftParen GetProof RightParen { extra.visit_get_proof() }
+    command ::= LeftParen GetProof RightParen { extra.0.visit_get_proof()? }
     //   ( get-unsat-assumptions )
-    command ::= LeftParen GetUnsatAssumptions RightParen { extra.visit_get_unsat_assumptions() }
+    command ::= LeftParen GetUnsatAssumptions RightParen { extra.0.visit_get_unsat_assumptions()? }
     //   ( get-unsat-core )
-    command ::= LeftParen GetUnsatCore RightParen { extra.visit_get_unsat_core() }
+    command ::= LeftParen GetUnsatCore RightParen { extra.0.visit_get_unsat_core()? }
     //   ( get-value )
-    command ::= LeftParen GetValue terms(xs) RightParen { extra.visit_get_value(xs) }
+    command ::= LeftParen GetValue terms(xs) RightParen { extra.0.visit_get_value(xs)? }
     //   ( pop ⟨numeral⟩ )
-    command ::= LeftParen Pop Numeral(x) RightParen { extra.visit_pop(x) }
+    command ::= LeftParen Pop Numeral(x) RightParen { extra.0.visit_pop(x)? }
     //   ( push ⟨numeral⟩ )
-    command ::= LeftParen Push Numeral(x) RightParen { extra.visit_push(x) }
+    command ::= LeftParen Push Numeral(x) RightParen { extra.0.visit_push(x)? }
     //   ( reset )
-    command ::= LeftParen Reset RightParen { extra.visit_reset() }
+    command ::= LeftParen Reset RightParen { extra.0.visit_reset()? }
     //   ( reset-assertions )
-    command ::= LeftParen ResetAssertions RightParen { extra.visit_reset_assertions() }
+    command ::= LeftParen ResetAssertions RightParen { extra.0.visit_reset_assertions()? }
     //   ( set-info ⟨attribute⟩ )
-    command ::= LeftParen SetInfo keyword(k) attribute_value(v) RightParen { extra.visit_set_info(k, v) }
+    command ::= LeftParen SetInfo keyword(k) attribute_value(v) RightParen { extra.0.visit_set_info(k, v)? }
     //   ( set-logic ⟨symbol⟩ )
-    command ::= LeftParen SetLogic bound_symbol(x) RightParen { extra.visit_set_logic(x) }
+    command ::= LeftParen SetLogic bound_symbol(x) RightParen { extra.0.visit_set_logic(x)? }
     //   ( set-option ⟨attribute⟩ )
-    command ::= LeftParen SetOption keyword(k) attribute_value(v) RightParen { extra.visit_set_option(k, v) }
+    command ::= LeftParen SetOption keyword(k) attribute_value(v) RightParen { extra.0.visit_set_option(k, v)? }
 }
 
 #[cfg(test)]
@@ -359,9 +365,10 @@ pub(crate) mod tests {
     use super::*;
     use crate::{concrete::*, lexer::Lexer};
 
-    pub(crate) fn parse_tokens<I: IntoIterator<Item = Token>>(tokens: I) -> Result<Command, ()> {
-        let mut builder = SyntaxBuilder;
-        let mut p = Parser::new(&mut builder);
+    pub(crate) fn parse_tokens<I: IntoIterator<Item = Token>>(tokens: I) -> Result<Command, Error> {
+        let mut builder = SyntaxBuilder::default();
+        let mut position = crate::Position::default();
+        let mut p = Parser::new((&mut builder, &mut position));
         for token in tokens.into_iter() {
             p.parse(token)?;
         }
@@ -502,9 +509,7 @@ pub(crate) mod tests {
 
         assert!(matches!(value, Command::DeclareDatatypes { .. }));
         // Test syntax visiting while we're at it.
-        assert_eq!(
-            value,
-            value.clone().accept(&mut crate::concrete::SyntaxBuilder)
-        );
+        let mut builder = crate::concrete::SyntaxBuilder::default();
+        assert_eq!(value, value.clone().accept(&mut builder).unwrap());
     }
 }
